/*	$NetBSD: bus_space.c,v 1.3 2019/12/14 02:58:20 tsutsui Exp $	*/

/*-
 * Copyright (c) 1998 The NetBSD Foundation, Inc.
 * All rights reserved.
 *
 * This code is derived from software contributed to The NetBSD Foundation
 * by Jason R. Thorpe of the Numerical Aerospace Simulation Facility,
 * NASA Ames Research Center.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY THE NETBSD FOUNDATION, INC. AND CONTRIBUTORS
 * ``AS IS'' AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED
 * TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR
 * PURPOSE ARE DISCLAIMED.  IN NO EVENT SHALL THE FOUNDATION OR CONTRIBUTORS
 * BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
 * CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
 * SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS
 * INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN
 * CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE)
 * ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE
 * POSSIBILITY OF SUCH DAMAGE.
 */

/*
 * Implementation of bus_space mapping for the DECstation.
 */

#include <sys/cdefs.h>
__KERNEL_RCSID(0, "$NetBSD: bus_space.c,v 1.3 2019/12/14 02:58:20 tsutsui Exp $");

#include <sys/param.h>
#include <sys/systm.h>

#include <machine/bus.h>

/* ARGSUSED */
int
bus_space_map(
	bus_space_tag_t t,
	bus_addr_t bpa,
	bus_size_t size,
	int flags,
	bus_space_handle_t *bshp)
{
	int cacheable = flags & BUS_SPACE_MAP_CACHEABLE;

	/* Mappings on the DECstation are always linear. */
	if (cacheable)
		*bshp = MIPS_PHYS_TO_KSEG0(bpa);
	else
		*bshp = MIPS_PHYS_TO_KSEG1(bpa);
	return (0);
}

/* ARGSUSED */
int
bus_space_alloc(
	bus_space_tag_t t,
	bus_addr_t rstart,
	bus_addr_t rend,
	bus_size_t size,
	bus_size_t alignment,
	bus_size_t boundary,
	int flags,
	bus_addr_t *bpap,
	bus_space_handle_t *bshp)
{

	/*
	 * Not meaningful on any currently-supported DECstation bus.
	 */
	return (EINVAL);
}

/* ARGSUSED */
void
bus_space_free(
	bus_space_tag_t t,
	bus_space_handle_t bsh,
	bus_size_t size)
{

	/*
	 * Not meaningful on any currently-supported DECstation bus.
	 */
	panic("bus_space_free: shouldn't be here");
}

void
bus_space_unmap(
	bus_space_tag_t t,
	bus_space_handle_t bsh,
	bus_size_t size)
{

	/* Nothing to do. */
}

/* ARGSUSED */
int
bus_space_subregion(
	bus_space_tag_t t,
	bus_space_handle_t bsh,
	bus_size_t offset,
	bus_size_t size,
	bus_space_handle_t *nbshp)
{

	*nbshp = bsh + offset;
	return (0);
}
